// hlavickove subory *******************************************************
// systemove funkcie
#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <string.h>
// GL a GLU kniznica
#include <GL/gl.h>
#include <GL/glu.h>

// .h subor
#include "glTerrain.h"

// konstruktor *************************************************************
glTerrain::glTerrain()
{
 all_triangles=0; cull_triangles=all_triangles;

 vertex=NULL; vertex1=NULL;
 normal=NULL; normal1=NULL;
 texcoords=NULL; texcoords1=NULL;
}

// destruktor **************************************************************
glTerrain::~glTerrain()
{
 if (texture_map.data!=NULL) free(texture_map.data);

 if (vertex!=NULL) free(vertex);
 if (vertex1!=NULL) free(vertex1);

 if (normal!=NULL) free(normal);
 if (normal1!=NULL) free(normal1);

 if (texcoords!=NULL) free(texcoords);
 if (texcoords1!=NULL) free(texcoords1);
}

int limit255(int a)
{
 if (a<0) return 0;
 else if (a>255) return 255;
      else return a;
}

float texture_factor(int h1,int h2)
{
 float t;
 t=(40-abs(h1-h2))/16.0f;

 if (t<0.0f) t=0.0f;
 else if (t>1.0f) t=1.0f;

 return t;
}

void GET_COLOR(TEXTURE *tex,int x,int y,unsigned char *r,
                                        unsigned char *g,
                                        unsigned char *b)
{
 if ((x<(int) tex->w)&&(y<(int) tex->h)) {
  *r=tex->data[(x+(y*tex->w))*3+0];
  *g=tex->data[(x+(y*tex->w))*3+1];
  *b=tex->data[(x+(y*tex->w))*3+2];
 }
}

void SET_COLOR(TEXTURE tex,int x,int y,unsigned char r,
                                       unsigned char g,
                                       unsigned char b)
{
 if ((x<(int) tex.w)&&(y<(int) tex.h)) {
  tex.data[(x+(y*tex.w))*3+0]=b;
  tex.data[(x+(y*tex.w))*3+1]=g;
  tex.data[(x+(y*tex.w))*3+2]=r;
 }
}

int glTerrain::texture_terrain(TEXTURE tt,unsigned char *hf)
{
 TEXTURE *tex[5];
 unsigned int x,y,xt,yt;
 float f0,f1,f2,f3,f4;
 unsigned int h1;
 unsigned char r[5],g[5],b[5];
 unsigned char new_r,new_g,new_b;

 tex[0]=NULL; tex[1]=NULL; tex[2]=NULL; tex[3]=NULL; tex[4]=NULL;

 char fname[256];
 strcpy(fname,"data/");
 strcat(fname,"terrain/");
 strcat(fname,"texture0_1.bmp\0");
 tex[0]=bmp_img.open_bitmap(fname);

 strcpy(fname,"data/");
 strcat(fname,"terrain/");
 strcat(fname,"texture0_2.bmp\0");
 tex[1]=bmp_img.open_bitmap(fname);

 strcpy(fname,"data/");
 strcat(fname,"terrain/");
 strcat(fname,"texture0_3.bmp\0");
 tex[2]=bmp_img.open_bitmap(fname);

 strcpy(fname,"data/");
 strcat(fname,"terrain/");
 strcat(fname,"texture0_4.bmp\0");
 tex[3]=bmp_img.open_bitmap(fname);

 strcpy(fname,"data/");
 strcat(fname,"terrain/");
 strcat(fname,"texture0_5.bmp\0");
 tex[4]=bmp_img.open_bitmap(fname);

 if ((tex[0]==NULL)||(tex[1]==NULL)||(tex[2]==NULL)||
     (tex[3]==NULL)||(tex[4]==NULL)) {
  fprintf(stderr,"Error opening terrain textures!\n");
  return(1);
 }

 for(y=0;y<tt.h;y++) {
  for(x=0;x<tt.w;x++) {
   //h1=limit255((tt.data[q+100*w]&0x0FFF)/11);
   h1=limit255(*(hf+3*x+3*1024*y));
     
   f4=texture_factor(256,h1);
   f3=texture_factor(192,h1);
   f2=texture_factor(128,h1);
   f1=texture_factor(64,h1);
   f0=texture_factor(0,h1);

   if (x<256) xt=x;
   else if (x<512) xt=x-256;
        else if (x<768) xt=x-512;
             else xt=x-768;
   if (y<256) yt=y;
   else if (y<512) yt=y-256;
        else if (y<768) yt=y-512;
             else yt=y-768;

   GET_COLOR(tex[0],xt,yt,&r[0],&g[0],&b[0]);
   GET_COLOR(tex[1],xt,yt,&r[1],&g[1],&b[1]);
   GET_COLOR(tex[2],xt,yt,&r[2],&g[2],&b[2]);
   GET_COLOR(tex[3],xt,yt,&r[3],&g[3],&b[3]);
   GET_COLOR(tex[4],xt,yt,&r[4],&g[4],&b[4]);

   new_r=(unsigned char) limit255((int) ((r[0]*f0)+(r[1]*f1)+(r[2]*f2)+(r[3]*f3)+(r[4]*f4)));
   new_g=(unsigned char) limit255((int) ((g[0]*f0)+(g[1]*f1)+(g[2]*f2)+(g[3]*f3)+(g[4]*f4)));
   new_b=(unsigned char) limit255((int) ((b[0]*f0)+(b[1]*f1)+(b[2]*f2)+(b[3]*f3)+(b[4]*f4)));

   SET_COLOR(tt,x,y,new_r,new_g,new_b);
  }
 }

 if (tex[4]!=NULL) bmp_img.close_bitmap(tex[4]);
 if (tex[3]!=NULL) bmp_img.close_bitmap(tex[3]);
 if (tex[2]!=NULL) bmp_img.close_bitmap(tex[2]);
 if (tex[1]!=NULL) bmp_img.close_bitmap(tex[1]);
 if (tex[0]!=NULL) bmp_img.close_bitmap(tex[0]);

 glGenTextures(1,&terrain_texture);
 // pripojenie textury
 glBindTexture(GL_TEXTURE_2D,terrain_texture);
 // nastavenie LINEAR filteringu
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
 // vygenerovanie textury
 glTexImage2D(GL_TEXTURE_2D,0,3,tt.w,tt.h,
              0,GL_BGR,GL_UNSIGNED_BYTE,tt.data);

 return(0);
}

void NormaliseNormal(GLfloat n1,GLfloat n2,GLfloat n3,
                     GLfloat *a,GLfloat *b,GLfloat *c)
{
 GLfloat d;
 d=n1; if (d<n2) d=n2; if (d<n3) d=n3;
 *a=n1/d; *b=n2/d; *c=n3/d;
}

void CalcNormal(GLfloat p1,GLfloat p2,GLfloat p3,
                GLfloat q1,GLfloat q2,GLfloat q3,
                GLfloat r1,GLfloat r2,GLfloat r3,
                GLfloat *a,GLfloat *b,GLfloat *c)
{
 GLfloat v11,v12,v13;
 GLfloat v21,v22,v23;
 GLfloat n1,n2,n3,nn1,nn2,nn3;

 v11=q1-p1; v12=q2-p2; v13=q3-p3;
 v21=r1-p1; v22=r2-p2; v23=r3-p3;

 n1=v12*v23-v13*v22;
 n2=v13*v21-v11*v23;
 n3=v11*v22-v12*v21;

 NormaliseNormal(n1,n2,n3,&nn1,&nn2,&nn3);

 *a=nn1; *b=nn2; *c=nn3;
}

void SmoothNormals(GLuint at,GLfloat *n)
{
 GLuint a,a1,a2,a3,a4,a5,a6;
 GLuint q,w,e;
 q=1024; w=32; e=q/w;

 for (a=12;a<(e*e*18);a+=18) {  // pouzit at!!!
  a1=a;
  a2=a1+12;
  a3=a2+9;
  a4=a1+(e*18-27);
  a5=a4+6;
  a6=a5+9;

  *(n+a1)=(*(n+a1)+*(n+a2)+*(n+a3)+*(n+a4)+*(n+a5)+*(n+a6))/6.0f;
  *(n+a1+1)=(*(n+a1+1)+*(n+a2+1)+*(n+a3+1)+*(n+a4+1)+*(n+a5+1)+*(n+a6+1))/6.0f;
  *(n+a1+2)=(*(n+a1+2)+*(n+a2+2)+*(n+a3+2)+*(n+a4+2)+*(n+a5+2)+*(n+a6+2))/6.0f;

  *(n+a2)=*(n+a1);
  *(n+a2+1)=*(n+a1+1);
  *(n+a2+2)=*(n+a1+2);

  *(n+a3)=*(n+a1);
  *(n+a3+1)=*(n+a1+1);
  *(n+a3+2)=*(n+a1+2);

  *(n+a4)=*(n+a1);
  *(n+a4+1)=*(n+a1+1);
  *(n+a4+2)=*(n+a1+2);

  *(n+a5)=*(n+a1);
  *(n+a5+1)=*(n+a1+1);
  *(n+a5+2)=*(n+a1+2);

  *(n+a6)=*(n+a1);
  *(n+a6+1)=*(n+a1+1);
  *(n+a6+2)=*(n+a1+2);
 }
}

// priprava terenu *********************************************************
int glTerrain::terrain_create(unsigned char *bmp_data)
{
 ta=0.0f; tb=0.0f;
 t1=0; t2=0; t3=0; t4=0;
 q=1024.0f; w=32.0f; e=q/w; r=3*1024;

 vertex=(GLfloat *) malloc(9*((GLuint) (e*e))*2*sizeof(GLfloat));
 if (!vertex) {
  fprintf(stderr,"Error allocating memory for vertexes!\n");
  return(1);
 }
 vertex1=(GLfloat *) malloc(9*((GLuint) (e*e))*2*sizeof(GLfloat));
 if (!vertex1) {
  fprintf(stderr,"Error allocating memory for vertexes1!\n");
  return(1);
 }

 normal=(GLfloat *) malloc(9*((GLuint) (e*e))*2*sizeof(GLfloat));
 if (!vertex) {
  fprintf(stderr,"Error allocating memory for normals!\n");
  return(1);
 }
 normal1=(GLfloat *) malloc(9*((GLuint) (e*e))*2*sizeof(GLfloat));
 if (!vertex1) {
  fprintf(stderr,"Error allocating memory for normals1!\n");
  return(1);
 }

 texcoords=(GLfloat *) malloc(6*((GLuint) (e*e))*2*sizeof(GLfloat));
 if (!texcoords) {
  fprintf(stderr,"Error allocating memory for texcoords!\n");
  return(1);
 }
 texcoords1=(GLfloat *) malloc(6*((GLuint) (e*e))*2*sizeof(GLfloat));
 if (!texcoords1) {
  fprintf(stderr,"Error allocating memory for texcoords1!\n");
  return(1);
 }

 a=0; b=0;
 GLfloat zf=1.0f;

 for (ty=0.0f;ty<(q-w);ty+=w) {
  for (tx=0.0f;tx<(q-w);tx+=w) {
   t1=(GLuint) (3*tx+(r*ty));
   t2=(GLuint) (3*(tx+w)+(r*ty));
   t3=(GLuint) (3*(tx+w)+(r*(ty+w)));
   t4=(GLuint) (3*tx+(r*(ty+w)));

   CalcNormal(zf*(tx),  (GLfloat) bmp_data[t1],zf*(-ty),
              zf*(tx+w),(GLfloat) bmp_data[t2],zf*(-ty),
              zf*(tx),  (GLfloat) bmp_data[t4],zf*(-ty-w),
              &n1,&n2,&n3);

   *(vertex+a)=tx; *(vertex+a+1)=(GLfloat) bmp_data[t1]; *(vertex+a+2)=-ty;
   *(normal+a)=n1; *(normal+a+1)=n2; *(normal+a+2)=n3; a+=3;
   *(texcoords+b)=tx/q; *(texcoords+b+1)=ty/q; b+=2;

   *(vertex+a)=tx+w; *(vertex+a+1)=(GLfloat) bmp_data[t2]; *(vertex+a+2)=-ty;
   *(normal+a)=n1; *(normal+a+1)=n2; *(normal+a+2)=n3; a+=3;
   *(texcoords+b)=(tx+w)/q; *(texcoords+b+1)=ty/q; b+=2;

   *(vertex+a)=tx; *(vertex+a+1)=(GLfloat) bmp_data[t4]; *(vertex+a+2)=-ty-w;
   *(normal+a)=n1; *(normal+a+1)=n2; *(normal+a+2)=n3; a+=3;
   *(texcoords+b)=tx/q; *(texcoords+b+1)=(ty+w)/q; b+=2;

   CalcNormal(zf*(tx+w),(GLfloat) bmp_data[t2],zf*(-ty),
              zf*(tx+w),(GLfloat) bmp_data[t3],zf*(-ty-w),
              zf*(tx),  (GLfloat) bmp_data[t4],zf*(-ty-w),
              &n1,&n2,&n3);

   *(vertex+a)=tx+w; *(vertex+a+1)=(GLfloat) bmp_data[t2]; *(vertex+a+2)=-ty;
   *(normal+a)=n1; *(normal+a+1)=n2; *(normal+a+2)=n3; a+=3;
   *(texcoords+b)=(tx+w)/q; *(texcoords+b+1)=ty/q; b+=2;

   *(vertex+a)=tx+w; *(vertex+a+1)=(GLfloat) bmp_data[t3]; *(vertex+a+2)=-ty-w;
   *(normal+a)=n1; *(normal+a+1)=n2; *(normal+a+2)=n3; a+=3;
   *(texcoords+b)=(tx+w)/q; *(texcoords+b+1)=(ty+w)/q; b+=2;

   *(vertex+a)=tx; *(vertex+a+1)=(GLfloat) bmp_data[t4]; *(vertex+a+2)=-ty-w;
   *(normal+a)=n1; *(normal+a+1)=n2; *(normal+a+2)=n3; a+=3;
   *(texcoords+b)=tx/q; *(texcoords+b+1)=(ty+w)/q; b+=2;

   ta+=1.0f/e;
  }
  ta=0.0f; tb+=1.0f/e;
 }

 all_triangles=a/9; cull_triangles=all_triangles;

 for (a=0;a<all_triangles*9;a+=9) {
  *(vertex+a)*=zf;   *(vertex+a+2)*=zf;
  *(vertex+a+3)*=zf; *(vertex+a+5)*=zf;
  *(vertex+a+6)*=zf; *(vertex+a+8)*=zf;

  *(vertex1+a)  =*(vertex+a);   *(vertex1+a+1)=*(vertex+a+1); *(vertex1+a+2)=*(vertex+a+2);
  *(vertex1+a+3)=*(vertex+a+3); *(vertex1+a+4)=*(vertex+a+4); *(vertex1+a+5)=*(vertex+a+5);
  *(vertex1+a+6)=*(vertex+a+6); *(vertex1+a+7)=*(vertex+a+7); *(vertex1+a+8)=*(vertex+a+8);
 }

 glVertexPointer(3,GL_FLOAT,0,vertex1);
 glEnableClientState(GL_VERTEX_ARRAY);
 //glNormalPointer(GL_FLOAT,0,normal1);
 //glEnableClientState(GL_NORMAL_ARRAY);
 ////glTexCoordPointer(2,GL_FLOAT,0,texcoords1);
 ////glEnableClientState(GL_TEXTURE_COORD_ARRAY);

 ////SmoothNormals(all_triangles,normal);

 texture_map.w=1024;
 texture_map.h=1024;
 texture_map.data=(unsigned char *) malloc(texture_map.w*texture_map.h*3);
 if (texture_map.data==NULL) {
  printf("Error allocating memory for terrain texture!");
  return (1);
 }
 ////if (texture_terrain(texture_map,bmp_data)) return(1);

 return(0);
}

// vratenie pola trojuholnikov *********************************************
GLfloat *glTerrain::terrain_getTriangles(void)
{
 return vertex;
}

// vratenie pola trojuholnikov *********************************************
GLfloat *glTerrain::terrain_getTriangles1(void)
{
 return vertex1;
}

// vratenie pola normal ****************************************************
GLfloat *glTerrain::terrain_getNormals(void)
{
 return normal;
}

// vratenie pola normal ****************************************************
GLfloat *glTerrain::terrain_getNormals1(void)
{
 return normal1;
}

// vratenie pola texcoords *************************************************
GLfloat *glTerrain::terrain_getTexCoords(void)
{
 return texcoords;
}

// vratenie pola texcoords1 ************************************************
GLfloat *glTerrain::terrain_getTexCoords1(void)
{
 return texcoords1;
}

// kreslenie terenu ********************************************************
void glTerrain::terrain_render(GLuint tris_num)
{
 glDrawArrays(GL_TRIANGLES,0,3*tris_num);
}

// kreslenie normal ********************************************************
void glTerrain::normals_render(GLuint tris_num)
{
 glDisable(GL_LIGHTING);
 glColor4f(1.0f,0.0f,0.0f,1.0f);
 glBegin(GL_LINES);
  for (a=0;a<tris_num*9;a+=9) {
   glVertex3fv(vertex1+a);
   glVertex3f(*(vertex1+a)+*(normal1+a)*20.0f,
              *(vertex1+a+1)+*(normal1+a+1)*20.0f,
              *(vertex1+a+2)+*(normal1+a+2)*20.0f);
  }
 glEnd();
 glEnable(GL_LIGHTING);
}

